from __future__ import division, absolute_import
import logging

import numpy as np
import matplotlib.pyplot as plt
from pylab import rcParams
from .utils import raster_to_image

logger = logging.getLogger(__name__)


def make_plot(source, reference, target,
              src_arr, ref_arr, tar_arr,
              output, bands):
    """ Create a diagnostic plot showing source, reference and matched image
    and the cumulative distribution functions for each band
    """
    rcParams['figure.figsize'] = 16, 14

    logger.debug("reading images")
    i1 = raster_to_image(source)
    i2 = raster_to_image(reference)
    i3 = raster_to_image(target)

    f, ((ax1, ax2, ax3), (ax4, ax5, ax6), (ax7, ax8, ax9)) = plt.subplots(3, 3)

    logger.debug("showing images")
    ax1.imshow(i1)
    ax1.set_title('Source')
    ax1.set_xticklabels([])
    ax1.set_yticklabels([])

    ax2.imshow(i2)
    ax2.set_title('Reference')
    ax2.set_xticklabels([])
    ax2.set_yticklabels([])

    ax3.imshow(i3)
    ax3.set_title('Matched to ' + ', '.join(str(b) for _, b in bands))
    ax3.set_xticklabels([])
    ax3.set_yticklabels([])

    logger.debug("RGB histograms")
    axes = (ax4, ax5, ax6)
    imgs = (i1, i2, i3)
    titles = ('Source', 'Reference', 'Output')
    bins = 32
    for i, axis in enumerate(axes):
        im = imgs[i]
        title = titles[i]
        # compressed for masked arrays, ravel for ndarray
        red, _ = np.histogram(im[:, :, 0].compressed(), bins, [0, 1])
        green, _ = np.histogram(im[:, :, 1].compressed(), bins, [0, 1])
        blue, _ = np.histogram(im[:, :, 2].compressed(), bins, [0, 1])
        for color, name in ((red, "red"), (green, "green"), (blue, "blue")):
            norm = color / im.size
            # axis.plot(norm, color=name, lw=2)
            axis.fill_between([float(x) / bins for x in range(bins)],
                              norm, facecolor=name, alpha=0.15)
        axis.set_title("{} RGB histogram".format(title))
        # axis.set_yticklabels([])
        axis.grid('on')

    logger.debug("CDF match plots")
    axes = (ax7, ax8, ax9)
    for b, band in bands:
        ax = axes[b]
        source_band = src_arr[b]
        reference_band = ref_arr[b]
        target_band = tar_arr[b]
        try:
            source_band = source_band.compressed()
        except:
            pass

        try:
            reference_band = reference_band.compressed()
        except:
            pass

        try:
            target_band = target_band.compressed()
        except:
            pass

        sv, sc = np.unique(source_band, return_counts=True)
        rv, rc = np.unique(reference_band, return_counts=True)
        tv, tc = np.unique(target_band, return_counts=True)

        scdf = np.cumsum(sc).astype(np.float64) / source_band.size
        rcdf = np.cumsum(rc).astype(np.float64) / reference_band.size
        tcdf = np.cumsum(tc).astype(np.float64) / target_band.size

        ax.set_title("{} cumulative distribution".format(band))
        ax.plot(sv, scdf, label="Source")
        ax.plot(rv, rcdf, label="Reference")
        ax.plot(tv, tcdf, '--r', lw=2, label="Match")
        if b == 1:
            ax.legend(loc=9, bbox_to_anchor=(0.5, -0.05))
        # ax.set_yticklabels([])
        ax.grid('on')

    plt.savefig(output, bbox_inches='tight')



import logging

import click
from rasterio.rio.options import creation_options
from rio_hist.match import hist_match_worker

logger = logging.getLogger('rio_hist')


@click.command('hist')
@click.option('--color-space', '-c', default="RGB",
              type=click.Choice(['RGB', 'LCH', 'Lab', 'LUV', 'XYZ']),
              help="Colorspace")
@click.option('--bands', '-b', default="1,2,3",
              help="comma-separated list of bands to match (default 1,2,3)")
@click.option('--plot', is_flag=True, default=False,
              help="create a <basename>_plot.png with diagnostic plots")
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.argument('src_path', type=click.Path(exists=True))
@click.argument('ref_path', type=click.Path(exists=True))
@click.argument('dst_path', type=click.Path(exists=False))
@click.pass_context
@creation_options
def hist(ctx, src_path, ref_path, dst_path,
         verbose, creation_options, bands, color_space, plot):
    """Color correction by histogram matching
    """
    if verbose:
        logger.setLevel(logging.DEBUG)

    hist_match_worker(src_path, ref_path, dst_path,
                      creation_options, bands, color_space, plot)



from __future__ import division, absolute_import
import warnings

import numpy as np
import rasterio
from rasterio.enums import ColorInterp, MaskFlags
from rio_color.colorspace import convert_arr, ColorSpace


def read_mask(src):
    """
    TODO Use RasterReader.dataset_mask in rasterio 0.36
    """
    masks = src.read_masks()

    # GDAL found dataset-wide alpha band or mask
    if src.mask_flags[0] & MaskFlags.per_dataset:
        return masks[0]

    # use Alpha mask if available and looks like RGB, even if nodata is shadowing
    elif src.count == 4 and src.colorinterp(1) == ColorInterp.red:
        return masks[3]

    # Or use the binary OR intersection of all GDALGetMaskBands
    else:
        mask = masks[0]
        for i in range(1, src.count):
            mask = mask | masks[i]
        return mask


def reshape_as_image(arr):
    """raster order (bands, rows, cols) -> image (rows, cols, bands)

    TODO Use rasterio.plot.reshape_as_image in rasterio 0.36?
    """
    return np.swapaxes(np.swapaxes(arr, 0, 2), 0, 1)


def reshape_as_raster(arr):
    """image order (rows, cols, bands) -> rasterio (bands, rows, cols)

    TODO Use rasterio.plot.reshape_as_image in rasterio 0.36?
    """
    return np.swapaxes(np.swapaxes(arr, 2, 0), 2, 1)


def cs_forward(arr, cs='rgb'):
    """ RGB (any dtype) to whatevs
    """
    arrnorm_raw = arr.astype('float64') / np.iinfo(arr.dtype).max
    arrnorm = arrnorm_raw[0:3]
    cs = cs.lower()
    if cs == 'rgb':
        return arrnorm
    elif cs == 'lch':
        return convert_arr(arrnorm,
                           src=ColorSpace.rgb,
                           dst=ColorSpace.lch)
    elif cs == 'lab':
        return convert_arr(arrnorm,
                           src=ColorSpace.rgb,
                           dst=ColorSpace.lab)
    elif cs == 'luv':
        return convert_arr(arrnorm,
                           src=ColorSpace.rgb,
                           dst=ColorSpace.luv)
    elif cs == 'xyz':
        return convert_arr(arrnorm,
                           src=ColorSpace.rgb,
                           dst=ColorSpace.xyz)


def cs_backward(arr, cs='rgb'):
    """ whatevs to RGB 8-bit
    """
    cs = cs.lower()
    if cs == 'rgb':
        return (arr * 255).astype('uint8')
    elif cs == 'lch':
        rgb = convert_arr(arr,
                          src=ColorSpace.lch,
                          dst=ColorSpace.rgb)
        return (rgb * 255).astype('uint8')
    elif cs == 'lab':
        rgb = convert_arr(arr,
                          src=ColorSpace.lab,
                          dst=ColorSpace.rgb)
        return (rgb * 255).astype('uint8')
    elif cs == 'luv':
        rgb = convert_arr(arr,
                          src=ColorSpace.luv,
                          dst=ColorSpace.rgb)
        return (rgb * 255).astype('uint8')
    elif cs == 'xyz':
        rgb = convert_arr(arr,
                          src=ColorSpace.xyz,
                          dst=ColorSpace.rgb)
        return (rgb * 255).astype('uint8')


def raster_to_image(raster):
    """Make an image-ordered 8bit 3-band array
    from a rasterio source
    """
    with rasterio.open(raster) as src:
        arr = src.read(masked=True)
    return reshape_as_image(cs_forward(arr, 'RGB'))



from __future__ import division, absolute_import
import logging
import os

import numpy as np
import rasterio
from rasterio.transform import guard_transform
from .utils import cs_forward, cs_backward, read_mask

logger = logging.getLogger(__name__)


def histogram_match(source, reference):
    """
    Adjust the values of a source array
    so that its histogram matches that of a reference array

    Parameters:
    -----------
        source: np.ndarray
        reference: np.ndarray

    Returns:
    -----------
        target: np.ndarray
            The output array with the same shape as source
            but adjusted so that its histogram matches the reference
    """
    orig_shape = source.shape
    source = source.ravel()

    if np.ma.is_masked(reference):
        logger.debug("ref is masked, compressing")
        reference = reference.compressed()
    else:
        logger.debug("ref is unmasked, raveling")
        reference = reference.ravel()

    # get the set of unique pixel values
    # and their corresponding indices and counts
    logger.debug("Get unique pixel values")
    s_values, s_idx, s_counts = np.unique(
        source, return_inverse=True, return_counts=True)
    r_values, r_counts = np.unique(reference, return_counts=True)
    s_size = source.size

    if np.ma.is_masked(source):
        logger.debug("source is masked; get mask_index and remove masked values")
        mask_index = np.ma.where(s_values.mask)
        s_size = np.ma.where(s_idx != mask_index[0])[0].size
        s_values = s_values.compressed()
        s_counts = np.delete(s_counts, mask_index)

    # take the cumsum of the counts; empirical cumulative distribution
    logger.debug("calculate cumulative distribution")
    s_quantiles = np.cumsum(s_counts).astype(np.float64) / s_size
    r_quantiles = np.cumsum(r_counts).astype(np.float64) / reference.size

    # find values in the reference corresponding to the quantiles in the source
    logger.debug("interpolate values from source to reference by cdf")
    interp_r_values = np.interp(s_quantiles, r_quantiles, r_values)

    if np.ma.is_masked(source):
        logger.debug("source is masked, add fill_value back at mask_index")
        interp_r_values = np.insert(interp_r_values, mask_index[0], source.fill_value)

    # using the inverted source indicies, pull out the interpolated pixel values
    logger.debug("create target array from interpolated values by index")
    target = interp_r_values[s_idx]

    if np.ma.is_masked(source):
        logger.debug("source is masked, remask those pixels by position index")
        target = np.ma.masked_where(s_idx == mask_index[0], target)
        target.fill_value = source.fill_value

    return target.reshape(orig_shape)


def calculate_mask(src, arr):
    msk = arr.mask
    if msk.sum() == 0:
        mask = None
        fill = None
    else:
        _gdal_mask = read_mask(src)
        mask = np.invert((_gdal_mask / 255).astype('bool'))
        fill = arr.fill_value
    return mask, fill


def hist_match_worker(src_path, ref_path, dst_path,
                      creation_options, bands, color_space, plot):
    """Match histogram of src to ref, outputing to dst
    optionally output a plot to <dst>_plot.png
    """
    logger.info("Matching {} to histogram of {} using {} color space".format(
        os.path.basename(src_path), os.path.basename(ref_path), color_space))

    with rasterio.open(src_path) as src:
        profile = src.profile.copy()
        src_arr = src.read(masked=True)
        src_mask, src_fill = calculate_mask(src, src_arr)
        src_arr = src_arr.filled()

    with rasterio.open(ref_path) as ref:
        ref_arr = ref.read(masked=True)
        ref_mask, ref_fill = calculate_mask(ref, ref_arr)
        ref_arr = ref_arr.filled()

    src = cs_forward(src_arr, color_space)
    ref = cs_forward(ref_arr, color_space)

    bixs = tuple([int(x) - 1 for x in bands.split(',')])
    band_names = [color_space[x] for x in bixs]  # assume 1 letter per band

    target = src.copy()
    for i, b in enumerate(bixs):
        logger.debug("Processing band {}".format(b))
        src_band = src[b]
        ref_band = ref[b]

        # Re-apply 2D mask to each band
        if src_mask is not None:
            logger.debug("apply src_mask to band {}".format(b))
            src_band = np.ma.asarray(src_band)
            src_band.mask = src_mask
            src_band.fill_value = src_fill

        if ref_mask is not None:
            logger.debug("apply ref_mask to band {}".format(b))
            ref_band = np.ma.asarray(ref_band)
            ref_band.mask = ref_mask
            ref_band.fill_value = ref_fill

        target[b] = histogram_match(src_band, ref_band)

    target_rgb = cs_backward(target, color_space)

    # re-apply src_mask to target_rgb and write ndv
    if src_mask is not None:
        logger.debug("apply src_mask to target_rgb")
        if not np.ma.is_masked(target_rgb):
            target_rgb = np.ma.asarray(target_rgb)
        target_rgb.mask = np.array((src_mask, src_mask, src_mask))
        target_rgb.fill_value = src_fill
        profile['count'] = 4
    else:
        profile['count'] = 3

    profile['dtype'] = 'uint8'
    profile['nodata'] = None
    profile['transform'] = guard_transform(profile['transform'])
    profile.update(creation_options)

    logger.info("Writing raster {}".format(dst_path))
    with rasterio.open(dst_path, 'w', **profile) as dst:
        dst.write(target_rgb[0], 1)
        dst.write(target_rgb[1], 2)
        dst.write(target_rgb[2], 3)
        if src_mask is not None:
            gdal_mask = (np.invert(src_mask) * 255).astype('uint8')
            dst.write(gdal_mask, 4)

    if plot:
        from .plot import make_plot
        outplot = os.path.splitext(dst_path)[0] + "_plot.png"
        logger.info("Writing figure to {}".format(outplot))
        make_plot(
            src_path, ref_path, dst_path,
            src, ref, target,
            output=outplot,
            bands=tuple(zip(bixs, band_names)))
